from loguru import logger
from .config import GenerationConfig
from tqdm import tqdm

try:
    from llama_cpp import Llama, llama_token_is_eog
    _GGUF_AVAILABLE = True
except:
    _GGUF_AVAILABLE = False

class GGUFModel:
    def __init__(
            self,
            model_path: str,
            n_gpu_layers: int = 0,
            max_seq_length: int = 4096,
            additional_model_config: dict = {}
    ) -> None:

        if not _GGUF_AVAILABLE:
            raise ImportError(
                "llama_cpp python module not found."
                "To use the GGUF model you must install llama cpp python manually."
            )

        additional_model_config["n_ctx"] = max_seq_length
        self.model = Llama(
            model_path=model_path,
            n_gpu_layers=n_gpu_layers,
            **additional_model_config
        )

    def generate(self, input_ids: list[int], config: GenerationConfig, stream: bool = False):
        if stream:
            return self._generate_stream(input_ids, config)
        return self._generate(input_ids, config)

    def _generate_stream(self, input_ids: list[int], config: GenerationConfig):
        size = 0
        input_size = len(input_ids)
        for token in self.model.generate(
            input_ids,
            temp=config.temperature,
            repeat_penalty=config.repetition_penalty,
            **config.additional_gen_config,
        ):
            yield token
            size += 1
            if (llama_token_is_eog(self.model._model.model, token) or 
                size + input_size >= config.max_length):
                break

    def _generate(self, input_ids: list[int], config: GenerationConfig) -> list:
        input_size = len(input_ids)
        tokens = []
        gen = tqdm(self.model.generate(
            input_ids,
            temp=config.temperature,
            repeat_penalty=config.repetition_penalty,
            **config.additional_gen_config,
        ))
        for token in gen:
            tokens.append(token)
            if (llama_token_is_eog(self.model._model.model, token) or 
                len(tokens) + input_size >= config.max_length):
                break
            gen.set_postfix({"tokens": input_size + len(tokens), "max tokens": config.max_length})

        return tokens


import torch
from dataclasses import dataclass, field

@dataclass
class HFModelConfig:
    model_path: str = "OuteAI/OuteTTS-0.3-1B"
    tokenizer_path: str = None 
    verbose: bool = False
    device: str = None
    dtype: torch.dtype = None
    additional_model_config: dict = field(default_factory=dict)
    wavtokenizer_model_path: str = None
    max_seq_length: int = 4096

@dataclass
class GGUFModelConfig(HFModelConfig):
    n_gpu_layers: int = 0

@dataclass
class EXL2ModelConfig(HFModelConfig):
    pass


__version__ = "0.3.0" 

from .interface import InterfaceHF, InterfaceGGUF, InterfaceEXL2, display_available_models
from .interface import HFModelConfig_v1, GGUFModelConfig_v1, EXL2ModelConfig_v1
from .interface import HFModelConfig_v2, GGUFModelConfig_v2, EXL2ModelConfig_v2
from .version.v1.alignment import CTCForcedAlignment
from .version.v2.interface import GenerationConfig


import torch
from loguru import logger
from transformers import AutoModelForCausalLM, AutoTokenizer
from .config import GenerationConfig

class HFModel:
    def __init__(
        self,
        model_path: str,
        device: str = None,
        dtype: torch.dtype = None,
        additional_model_config: dict = {}
    ) -> None:
        self.device = torch.device(
            device if device is not None
            else "cuda" if torch.cuda.is_available()
            else "cpu"
        )
        self.dtype = dtype
        self.model = AutoModelForCausalLM.from_pretrained(
            model_path,
            torch_dtype=dtype,
            **additional_model_config
        ).to(self.device)

    def generate(self, input_ids: torch.Tensor, config: GenerationConfig, stream: bool = False) -> list[int]:
        if stream:
            raise NotImplementedError("Stream generation is not supported for HF models.")

        return self.model.generate(
            input_ids,
            max_length=config.max_length,
            temperature=config.temperature,
            repetition_penalty=config.repetition_penalty,
            do_sample=True,
            **config.additional_gen_config,
        )[0].tolist()


import torch
import torchaudio
from dataclasses import dataclass, field
from loguru import logger
import os
import json
import threading
import queue
import re
import polars as pl
from pathlib import Path

from ...wav_tokenizer.audio_codec import AudioCodec
from .prompt_processor import PromptProcessor
from ...models.hf_model import HFModel
from ...models.gguf_model import GGUFModel
from ...models.exl2_model import EXL2Model
from ...models.config import GenerationConfig
from ...whisper import transcribe
from .config import HFModelConfig, GGUFModelConfig, EXL2ModelConfig
from ..playback import ModelOutput
from .alignment import CTCForcedAlignment

_BASE_DIR = os.path.dirname(__file__)
_DEFAULT_SPEAKERS_DIR = os.path.join(_BASE_DIR, "default_speakers/parquet/speakers.parquet")

if Path(_DEFAULT_SPEAKERS_DIR).exists():
    _DEFAULT_SPEAKERS = {i['speaker']: i for i in pl.read_parquet(_DEFAULT_SPEAKERS_DIR).to_dicts()}
else:
    _DEFAULT_SPEAKERS = None

class InterfaceHF:
    def __init__(
        self,
        config: HFModelConfig
    ) -> None:
        self.config = config
        self.verbose = config.verbose
        self.audio_codec = AudioCodec(config.device, config.wavtokenizer_model_path)
        self.prompt_processor = PromptProcessor(config.tokenizer_path)
        self.model = self.get_model()
        
    def get_model(self):
        return HFModel(
            self.config.model_path, 
            self.config.device, 
            self.config.dtype, 
            self.config.additional_model_config
        )

    def _prepare_prompt(self, prompt: str):
        return self.prompt_processor.tokenizer.encode(
            prompt,
            add_special_tokens=False,
            return_tensors="pt"
        ).to(self.model.device)

    def check_input_prompt_size(self, text: str):
        input_ids = self.prompt_processor.tokenizer.encode(text, add_special_tokens=False)
        if len(input_ids) >= 1500:
            logger.warning(
            f"Input prompt size exceeds 1500 tokens (current size: {len(input_ids)} tokens). "
            "This may reduce the effective context window for the model and impact generation accuracy.\n"
            "To achieve the best results, aim for the generation window to fall within a 1-30 second range.\n"
            "If you experience issues with the model's output, consider using a shorter speaker reference or reducing the length of your input text."
        )

    def prepare_prompt(self, text: str, voice_characteristics: str = None, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, voice_characteristics, speaker)
        self.check_input_prompt_size(prompt)
        return self._prepare_prompt(prompt)

    def get_audio(self, tokens):
        output = self.prompt_processor.extract_audio_from_tokens(tokens)
        if not output:
            logger.warning("No audio tokens found in the output")
            return None

        return self.audio_codec.decode(
            torch.tensor([[output]], dtype=torch.int64).to(self.audio_codec.device)
        )

    def create_speaker(
            self,
            audio_path: str,
            transcript: str = None,
            whisper_model: str = "turbo",
            whisper_device = None,
            free_ctc_model: bool = True
        ):
        if transcript is None:
            logger.info("Transcription not provided, transcribing audio with whisper.")
            transcript = transcribe.transcribe_once(
                audio_path=audio_path,
                model=whisper_model,
                device=whisper_device
            )
        if not transcript: 
            raise ValueError("Transcript text is empty")

        if not hasattr(self, 'ctc') or not isinstance(self.ctc, CTCForcedAlignment):
            self.ctc = CTCForcedAlignment(self.config.device)

        words = self.ctc.align(audio_path, transcript)
        full_codes = self.audio_codec.encode(
            self.audio_codec.convert_audio_tensor(
                audio=torch.cat([i["audio"] for i in words], dim=1),
                sr=self.ctc.sample_rate
            ).to(self.audio_codec.device)
        ).tolist()

        data = []
        start = 0
        for i in words:
            end = int(round((i["x1"] / self.ctc.sample_rate) * 75))
            word_tokens = full_codes[0][0][start:end]
            start = end
            if not word_tokens:
                word_tokens = [1]

            data.append({
                "word": i["word"],
                "duration": round(len(word_tokens) / 75, 2),
                "codes": word_tokens
            })

        if free_ctc_model:
            self.ctc.free()
            del self.ctc

        return {
            "text": transcript,
            "words": data,
        }

    def save_speaker(self, speaker: dict, path: str):
        with open(path, "w") as f: json.dump(speaker, f, indent=2)

    def load_speaker(self, path: str):
        with open(path, "r") as f: return json.load(f)

    def print_default_speakers(self):
        if _DEFAULT_SPEAKERS is None:
            logger.error(f"There was an issue loading default speakers.")
        else:
            logger.info(f"Available default speakers v2: {list(_DEFAULT_SPEAKERS.keys())}")
    
    def load_default_speaker(self, name: str):
        name = name.lower().strip()
        if _DEFAULT_SPEAKERS is None:
            raise ValueError(f"There was an issue loading default speakers.")

        if name not in _DEFAULT_SPEAKERS:
            raise ValueError(f"Speaker {name} not found for language {list(_DEFAULT_SPEAKERS.keys())}")

        return _DEFAULT_SPEAKERS.get(name, {})

    def check_generation_max_length(self, max_length):
        if max_length is None:
            raise ValueError("max_length must be specified.")
        if max_length > self.config.max_seq_length:
            raise ValueError(f"Requested max_length ({max_length}) exceeds the current max_seq_length ({self.config.max_seq_length}).")

    def _generate(self, input_ids, config: GenerationConfig):
        output = self.model.generate(
            input_ids=input_ids,
            config=config
        )
        return output[input_ids.size()[-1]:]
    
    def generate(self, config: GenerationConfig) -> ModelOutput:
        self.check_generation_max_length(config.max_length)
        if config.text is None:
            raise ValueError("text can not be empty!")
        if config.voice_characteristics is not None:
            logger.warning("Voice characteristics is an experimental feature that may not have any effect on the output. The model might process these settings differently than expected.")
        input_ids = self.prepare_prompt(config.text, config.voice_characteristics, config.speaker)
        output = self._generate(input_ids, config)
        audio = self.get_audio(output)
        return ModelOutput(audio, self.audio_codec.sr)

class InterfaceGGUF(InterfaceHF):
    def __init__(self, config: GGUFModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return GGUFModel(
            model_path=self.config.model_path,
            n_gpu_layers=self.config.n_gpu_layers,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config
        )

    def _generate(self, input_ids, config):
        return self.model.generate(
            input_ids=input_ids,
            config=config,
        )

    def _prepare_prompt(self, prompt: str):
        return self.prompt_processor.tokenizer.encode(prompt, add_special_tokens=False)

class InterfaceEXL2(InterfaceHF):
    def __init__(self, config: EXL2ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return EXL2Model(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate(self, input_ids, config):
        return self.model.generate(
            input_ids=input_ids,
            config=config,
            additional_dynamic_generator_config=config.additional_dynamic_generator_config
        )






import torch
from typing import Union
from loguru import logger
from typing import Union, overload

from .version.v1.interface import InterfaceHF as _InterfaceHF_v1
from .version.v1.interface import InterfaceGGUF as _InterfaceGGUF_v1
from .version.v1.interface import InterfaceEXL2 as _InterfaceEXL2_v1
from .version.v1.interface import HFModelConfig as HFModelConfig_v1
from .version.v1.interface import GGUFModelConfig as GGUFModelConfig_v1
from .version.v1.interface import EXL2ModelConfig as EXL2ModelConfig_v1

from .version.v2.interface import InterfaceHF as _InterfaceHF_v2
from .version.v2.interface import InterfaceGGUF as _InterfaceGGUF_v2
from .version.v2.interface import InterfaceEXL2 as _InterfaceEXL2_v2
from .version.v2.interface import HFModelConfig as HFModelConfig_v2
from .version.v2.interface import GGUFModelConfig as GGUFModelConfig_v2
from .version.v2.interface import EXL2ModelConfig as EXL2ModelConfig_v2

MODEL_CONFIGS = {
    "0.1": {
        "tokenizer": "OuteAI/OuteTTS-0.1-350M",
        "sizes": ["350M"],
        "links": ["https://huggingface.co/OuteAI/OuteTTS-0.1-350M", "https://huggingface.co/OuteAI/OuteTTS-0.1-350M-GGUF"],
        "languages": ["en"],
        "hf_interface": _InterfaceHF_v1,
        "gguf_interface": _InterfaceGGUF_v1,
        "max_seq_length": 4096
    },
    "0.2": {
        "tokenizer": "OuteAI/OuteTTS-0.2-500M",
        "sizes": ["500M"],
        "links": ["https://huggingface.co/OuteAI/OuteTTS-0.2-500M", "https://huggingface.co/OuteAI/OuteTTS-0.2-500M-GGUF"],
        "languages": ["en", "ja", "ko", "zh"],
        "hf_interface": _InterfaceHF_v1,
        "gguf_interface": _InterfaceGGUF_v1,
        "exl2_interface": _InterfaceEXL2_v1,
        "max_seq_length": 4096
    },
    "0.3": {
        "tokenizer": "OuteAI/OuteTTS-0.3-500M",
        "sizes": ["1B", "500M"],

        "links": ["https://huggingface.co/OuteAI/OuteTTS-0.3-500M", "https://huggingface.co/OuteAI/OuteTTS-0.3-500M-GGUF",
                  "https://huggingface.co/OuteAI/OuteTTS-0.3-1B", "https://huggingface.co/OuteAI/OuteTTS-0.3-1B-GGUF"],

        "languages": ["en", "ja", "ko", "zh", "de", "fr"],
        "hf_interface": _InterfaceHF_v2,
        "gguf_interface": _InterfaceGGUF_v2,
        "exl2_interface": _InterfaceEXL2_v2,
        "max_seq_length": 4096
    },
}

def display_available_models():
    print("\n=== Available OuteTTS Models ===\n")
    separator = "-" * 50
    for version, details in MODEL_CONFIGS.items():
        print(separator)
        print(f"Version: {version}")
        print(f"Supported Languages: {', '.join(details['languages'])}")
        print(f"Model Sizes: {', '.join(details['sizes'])}")
        print(f"Available Formats: HF, GGUF")
        print(f"Tokenizer: {details['tokenizer']}")
        print(f"Links: {', '.join(details['links'])}")
        print(separator + "\n")

def get_model_config(version: str):
    """
    Retrieve the configuration for a given model version.
    """
    if version not in MODEL_CONFIGS:
        raise ValueError(f"Unsupported model version '{version}'. Supported versions are: {list(MODEL_CONFIGS.keys())}")
    return MODEL_CONFIGS[version]

def check_max_length(max_seq_length: int, model_max_seq_length: int):
    if max_seq_length is None:
        raise ValueError("max_seq_length must be specified.")
    if max_seq_length > model_max_seq_length:
        raise ValueError(f"Requested max_seq_length ({max_seq_length}) exceeds the maximum supported length ({model_max_seq_length}).")
    
def InterfaceHF(
        model_version: str,
        cfg: Union[HFModelConfig_v1, HFModelConfig_v2]
    ) -> Union[_InterfaceHF_v1, _InterfaceHF_v2]:
    """
    Creates and returns a Hugging Face model interface for OuteTTS.

    Parameters
    ----------
    model_version : str
        Version identifier for the model to be loaded
    cfg : HFModelConfig_v1 | HFModelConfig_v2
        Configuration object containing parameters

    Returns
    -------
    An instance of interface based on the specified version.
    """

    config = get_model_config(model_version)
    cfg.tokenizer_path = cfg.tokenizer_path or config["tokenizer"]

    if model_version in ["0.1", "0.2"]:
        languages = config["languages"]
        if cfg.language not in languages:
            raise ValueError(f"Language '{cfg.language}' is not supported by model version '{model_version}'. Supported languages are: {languages}")
        cfg.languages = languages

    interface_class = config["hf_interface"]

    check_max_length(cfg.max_seq_length, config["max_seq_length"])

    return interface_class(cfg)

def InterfaceGGUF(
        model_version: str,
        cfg: Union[GGUFModelConfig_v1, GGUFModelConfig_v2]
    ) -> Union[_InterfaceGGUF_v1, _InterfaceGGUF_v2]:
    """
    Creates and returns a GGUF model interface for OuteTTS.

    Parameters
    ----------
    model_version : str
        Version identifier for the model to be loaded
    cfg : GGUFModelConfig_v1 | GGUFModelConfig_v2
        Configuration object containing parameters

    Returns
    -------
    An instance of interface based on the specified version.
    """

    if not cfg.model_path.lower().endswith('.gguf'):
        raise ValueError(f"Model path must point to a .gguf file, got: '{cfg.model_path}'")

    config = get_model_config(model_version)
    cfg.tokenizer_path = cfg.tokenizer_path or config["tokenizer"]

    if model_version in ["0.1", "0.2"]:
        languages = config["languages"]
        if cfg.language not in languages:
            raise ValueError(f"Language '{cfg.language}' is not supported by model version '{model_version}'. Supported languages are: {languages}")
        cfg.languages = languages

    check_max_length(cfg.max_seq_length, config["max_seq_length"])

    interface_class = config["gguf_interface"]
    return interface_class(cfg)

def InterfaceEXL2(
        model_version: str,
        cfg: Union[EXL2ModelConfig_v1, EXL2ModelConfig_v2]
    ) -> Union[_InterfaceEXL2_v1, _InterfaceEXL2_v2]:
    """
    Creates and returns a GGUF model interface for OuteTTS.

    Parameters
    ----------
    model_version : str
        Version identifier for the model to be loaded
    cfg : EXL2ModelConfig_v1 | EXL2ModelConfig_v2
        Configuration object containing parameters

    Returns
    -------
    An instance of interface based on the specified version.
    """

    config = get_model_config(model_version)
    cfg.tokenizer_path = cfg.tokenizer_path or config["tokenizer"]

    if model_version in ["0.1", "0.2"]:
        languages = config["languages"]
        if cfg.language not in languages:
            raise ValueError(f"Language '{cfg.language}' is not supported by model version '{model_version}'. Supported languages are: {languages}")
        cfg.languages = languages

    check_max_length(cfg.max_seq_length, config["max_seq_length"])

    interface_class = config["exl2_interface"]
    return interface_class(cfg)



from loguru import logger
from .config import GenerationConfig

try:
    from exllamav2 import ExLlamaV2, ExLlamaV2Config, ExLlamaV2Cache, ExLlamaV2Tokenizer
    from exllamav2.generator import ExLlamaV2DynamicGenerator, ExLlamaV2DynamicJob, ExLlamaV2Sampler
    _EXL2_AVAILABLE = True
except:
    _EXL2_AVAILABLE = False

class EXL2Model:
    def __init__(
            self,
            model_path: str,
            max_seq_length: int,
            additional_model_config: dict = {},
    ) -> None:

        if not _EXL2_AVAILABLE:
            raise ImportError(
                "exllamav2 python module not found."
                "To use the EXL2 model you must install exllamav2 manually."
            )

        config = ExLlamaV2Config(model_path)
        config.arch_compat_overrides()
        self.model = ExLlamaV2(config)
        self.cache = ExLlamaV2Cache(self.model, max_seq_len=config.max_seq_len, lazy=True)
        self.model.load_autosplit(self.cache, progress=True)
        self.tokenizer = ExLlamaV2Tokenizer(config)

    def generate(self, input_ids: str, config: GenerationConfig, additional_dynamic_generator_config: dict, stream: bool = False) -> list[int]:
        generator = ExLlamaV2DynamicGenerator(
            model = self.model,
            cache = self.cache,
            tokenizer = self.tokenizer,
            **additional_dynamic_generator_config
        )
        if stream:
            raise NotImplementedError("Stream generation is not supported for EXL2 models.")

        gen_settings = ExLlamaV2Sampler.Settings(
                    token_repetition_penalty=config.repetition_penalty,
                    temperature=config.temperature,
                    **config.additional_gen_config
                ),

        input_size = self.tokenizer.encode(input_ids).size()[-1]

        output = generator.generate(
            prompt = input_ids,
            max_new_tokens = config.max_length,
            add_bos = False,
            decode_special_tokens=True,
            gen_settings = gen_settings,
            stop_conditions=[self.tokenizer.eos_token_id]
        )

        return self.tokenizer.encode(output).flatten().tolist()[input_size:]



from ...wav_tokenizer.audio_codec import AudioCodec
from .prompt_processor import PromptProcessor
from ...models.hf_model import HFModel
from ...models.gguf_model import GGUFModel
from ...models.exl2_model import EXL2Model
from ...models.config import GenerationConfig
from ...whisper import transcribe
from ..playback import ModelOutput
import torch
from .alignment import CTCForcedAlignment
import torchaudio
from dataclasses import dataclass, field
from loguru import logger
import os
import json
import threading
import queue
import re

BASE_DIR = os.path.dirname(__file__)

DEFAULT_SPEAKERS_DIR = os.path.join(BASE_DIR, "default_speakers")

def get_speaker_path(speaker_name):
    return os.path.join(DEFAULT_SPEAKERS_DIR, f"{speaker_name}.json")

_DEFAULT_SPEAKERS = {
    "en": {
        "male_1": get_speaker_path("en_male_1"),
        "male_2": get_speaker_path("en_male_2"),
        "male_3": get_speaker_path("en_male_3"),
        "male_4": get_speaker_path("en_male_4"),
        "female_1": get_speaker_path("en_female_1"),
        "female_2": get_speaker_path("en_female_2"),
    },
    "ja": {
        "male_1": get_speaker_path("ja_male_1"),
        "female_1": get_speaker_path("ja_female_1"),
        "female_2": get_speaker_path("ja_female_2"),
        "female_3": get_speaker_path("ja_female_3"),
    },
    "ko": {
        "male_1": get_speaker_path("ko_male_1"),
        "male_2": get_speaker_path("ko_male_2"),
        "female_1": get_speaker_path("ko_female_1"),
        "female_2": get_speaker_path("ko_female_2"),
    },
    "zh": {
        "male_1": get_speaker_path("zh_male_1"),
        "female_1": get_speaker_path("zh_female_1"),
    }
}

@dataclass
class HFModelConfig:
    model_path: str = "OuteAI/OuteTTS-0.2-500M"
    language: str = "en"
    tokenizer_path: str = None
    languages: list = field(default_factory=list)
    verbose: bool = False
    device: str = None
    dtype: torch.dtype = None
    additional_model_config: dict = field(default_factory=dict)
    wavtokenizer_model_path: str = None
    max_seq_length: int = 4096

@dataclass
class GGUFModelConfig(HFModelConfig):
    n_gpu_layers: int = 0

@dataclass
class EXL2ModelConfig(HFModelConfig):
    pass

class InterfaceHF:
    def __init__(
        self,
        config: HFModelConfig
    ) -> None:
        self.device = torch.device(
            config.device if config.device is not None
            else "cuda" if torch.cuda.is_available()
            else "cpu"
        )
        self.config = config
        self._device = config.device
        self.languages = config.languages
        self.language = config.language
        self.verbose = config.verbose

        self.audio_codec = AudioCodec(self.device, config.wavtokenizer_model_path)
        self.prompt_processor = PromptProcessor(config.tokenizer_path, self.languages)
        self.model = HFModel(config.model_path, self.device, config.dtype, config.additional_model_config)

    def prepare_prompt(self, text: str, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, self.language, speaker)
        return self.prompt_processor.tokenizer.encode(
            prompt,
            add_special_tokens=False,
            return_tensors="pt"
        ).to(self.model.device)

    def get_audio(self, tokens):
        output = self.prompt_processor.extract_audio_from_tokens(tokens)
        if not output:
            logger.warning("No audio tokens found in the output")
            return None

        return self.audio_codec.decode(
            torch.tensor([[output]], dtype=torch.int64).to(self.audio_codec.device)
        )

    def create_speaker(
            self,
            audio_path: str,
            transcript: str = None,
            whisper_model: str = "turbo",
            whisper_device = None
        ):

        if transcript is None:
            logger.info("Transcription not provided, transcribing audio with whisper.")
            transcript = transcribe.transcribe_once(
                audio_path=audio_path,
                model=whisper_model,
                device=whisper_device
            )

        if not transcript:
            raise ValueError("Transcript text is empty")

        ctc = CTCForcedAlignment(self.languages, self._device)
        words = ctc.align(audio_path, transcript, self.language)
        ctc.free()

        full_codes = self.audio_codec.encode(
            self.audio_codec.convert_audio_tensor(
                audio=torch.cat([i["audio"] for i in words], dim=1),
                sr=ctc.sample_rate
            ).to(self.audio_codec.device)
        ).tolist()

        data = []
        start = 0
        for i in words:
            end = int(round((i["x1"] / ctc.sample_rate) * 75))
            word_tokens = full_codes[0][0][start:end]
            start = end
            if not word_tokens:
                word_tokens = [1]

            data.append({
                "word": i["word"],
                "duration": round(len(word_tokens) / 75, 2),
                "codes": word_tokens
            })

        return {
            "text": transcript,
            "words": data,
            "language": self.language
        }

    def save_speaker(self, speaker: dict, path: str):
        with open(path, "w") as f:
            json.dump(speaker, f, indent=2)

    def load_speaker(self, path: str):
        with open(path, "r") as f:
            return json.load(f)

    def print_default_speakers(self):
        total_speakers = sum(len(speakers) for speakers in _DEFAULT_SPEAKERS.values())
        print("\n=== ALL AVAILABLE SPEAKERS ===")
        print(f"Total: {total_speakers} speakers across {len(_DEFAULT_SPEAKERS)} languages")
        print("-" * 50)

        for language, speakers in _DEFAULT_SPEAKERS.items():
            print(f"\n{language.upper()} ({len(speakers)} speakers):")
            for speaker_name in speakers.keys():
                print(f"  - {speaker_name}")

        print("\n\n=== SPEAKERS FOR CURRENT INTERFACE LANGUAGE ===")
        if self.language.lower() in _DEFAULT_SPEAKERS:
            current_speakers = _DEFAULT_SPEAKERS[self.language.lower()]
            print(f"Language: {self.language.upper()} ({len(current_speakers)} speakers)")
            print("-" * 50)
            for speaker_name in current_speakers.keys():
                print(f"  - {speaker_name}")
        else:
            print(f"No speakers available for current language: {self.language}")

        print("\nTo use a speaker: load_default_speaker(name)\n")

    def load_default_speaker(self, name: str):
        name = name.lower().strip()
        language = self.language.lower().strip()
        if language not in _DEFAULT_SPEAKERS:
            raise ValueError(f"Speaker for language {language} not found")

        speakers = _DEFAULT_SPEAKERS[language]
        if name not in speakers:
            raise ValueError(f"Speaker {name} not found for language {language}")

        return self.load_speaker(speakers[name])

    def change_language(self, language: str):
        language = language.lower().strip()
        if language not in self.languages:
            raise ValueError(f"Language {language} is not supported by the current model")
        self.language = language

    def check_generation_max_length(self, max_length):
        if max_length is None:
            raise ValueError("max_length must be specified.")
        if max_length > self.config.max_seq_length:
            raise ValueError(f"Requested max_length ({max_length}) exceeds the current max_seq_length ({self.config.max_seq_length}).")

    def generate(
            self,
            text: str,
            speaker: dict = None,
            temperature: float = 0.1,
            repetition_penalty: float = 1.1,
            max_length: int = 4096,
            additional_gen_config={},
        ) -> ModelOutput:
        input_ids = self.prepare_prompt(text, speaker)
        if self.verbose:
            logger.info(f"Input tokens: {input_ids.size()[-1]}")
            logger.info("Generating audio...")

        self.check_generation_max_length(max_length)

        output = self.model.generate(
            input_ids=input_ids,
            config=GenerationConfig(
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                max_length=max_length,
                additional_gen_config=additional_gen_config,
            )
        )
        audio = self.get_audio(output[input_ids.size()[-1]:])
        if self.verbose:
            logger.info("Audio generation completed")

        return ModelOutput(audio, self.audio_codec.sr)

    # ------------------------------------------------ #
    # Audio Streaming Implementation
    # In development
    # Issues:
    # - Audio popping occurs when starting new chunk
    # - Needs smoother chunk transitions
    # ------------------------------------------------ #

    def _create_audio_chunk(self, tokens: list[int], idx: int):
        audio = self.get_audio(tokens)
        size = audio.size()
        audio = audio[:, idx:]
        return ModelOutput(audio, self.audio_codec.sr), size[-1]

    def _generate_stream(
        self,
        text: str,
        speaker: dict = None,
        temperature: float = 0.1,
        repetition_penalty: float = 1.1,
        max_length: int = 4096,
        additional_gen_config: dict = {},
        additional_dynamic_generator_config: dict = {},
        chunk_size: int = 8,
        stream_segments: bool = False
    ):
        if chunk_size < 4:
            raise ValueError("chunk_size must be at least 4.")

        self.check_generation_max_length(max_length)

        code_end_token = self.prompt_processor.tokenizer.encode(
            self.prompt_processor.special_tokens["code_end"],
            add_special_tokens=False
        )[0]

        input_ids = self.prepare_prompt(text, speaker)

        audio_buffer = []
        start_index = 0
        chunks = 0

        gen_config = {
            "input_ids": input_ids,
            "config": GenerationConfig(
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                max_length=max_length,
                additional_gen_config=additional_gen_config,
            ),
            "stream": True
        }
        # if additional_dynamic_generator_config:
            # gen_config["additional_dynamic_generator_config"] = additional_dynamic_generator_config

        for token in self.model.generate(**gen_config):

            audio_buffer.append(token)

            if not stream_segments:

                if token == code_end_token:
                    chunks += 1

                if chunks == chunk_size:
                    output, start_index = self._create_audio_chunk(audio_buffer, start_index)
                    chunks = 0
                    yield output

        if audio_buffer:
            output, start_index = self._create_audio_chunk(audio_buffer, start_index)
            chunks = 0
            yield output

    def _audio_player(self):
        while True:
            chunk = self.audio_queue.get()
            if chunk is None:
                # No more audio chunks to process
                break
            chunk.play()
            self.audio_queue.task_done()

    def split_text(self, text, delimiters=None):
        if delimiters is None:
            delimiters = ['.', '?', '!',
                        ';', ':', '\n',
                        '\r\n', '\r',
                        '\t',
                        '。', '？', '！',
                        '…']
        pattern = '|'.join(map(re.escape, delimiters))
        split_result = re.split(pattern, text)
        cleaned_sentences = [sentence.strip() for sentence in split_result]
        cleaned_sentences = [s for s in cleaned_sentences if s]
        return cleaned_sentences

    def generate_stream(
        self,
        text: str,
        speaker: dict = None,
        temperature: float = 0.1,
        repetition_penalty: float = 1.1,
        max_length: int = 4096,
        additional_gen_config: dict = {},
        additional_dynamic_generator_config: dict = {},
        chunk_size: int = 8,
        stream_segments: bool = True
    ):
        self.audio_queue = queue.Queue()
        audio_thread = threading.Thread(target=self._audio_player)
        audio_thread.start()

        logger.info(f"Processing text | Length: {len(text)} chars")
        segments = self.split_text(text)
        logger.info(f"Split text into {len(segments)} segments | Avg length: {sum(len(s) for s in segments)/len(segments):.1f} chars")

        if stream_segments:
            logger.info("stream_segments is set to True, chunk_size will be ignored.")

        try:
            for idx, segment in enumerate(segments):
                logger.info(f"Processing segment {idx+1}/{len(segments)}")
                for i in self._generate_stream(
                    text=segment,
                    speaker=speaker,
                    temperature=temperature,
                    repetition_penalty=repetition_penalty,
                    max_length=max_length,
                    additional_gen_config=additional_gen_config,
                    additional_dynamic_generator_config=additional_dynamic_generator_config,
                    chunk_size=chunk_size,
                    stream_segments=stream_segments
                ):
                    self.audio_queue.put(i)
        except Exception as e:
            logger.error(e)

        self.audio_queue.put(None)
        audio_thread.join()
        del self.audio_queue

class InterfaceGGUF(InterfaceHF):
    def __init__(
        self,
        config: GGUFModelConfig
    ) -> None:
        self.device = torch.device(
            config.device if config.device is not None
            else "cuda" if torch.cuda.is_available()
            else "cpu"
        )
        self.config = config
        self._device = config.device
        self.languages = config.languages
        self.language = config.language
        self.verbose = config.verbose

        self.audio_codec = AudioCodec(self.device, config.wavtokenizer_model_path)
        self.prompt_processor = PromptProcessor(config.tokenizer_path, self.languages)
        self.model = GGUFModel(
            model_path=config.model_path,
            n_gpu_layers=config.n_gpu_layers,
            max_seq_length=config.max_seq_length,
            additional_model_config=config.additional_model_config
        )

    def prepare_prompt(self, text: str, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, self.language, speaker)
        return self.prompt_processor.tokenizer.encode(prompt, add_special_tokens=False)

    def generate(
            self,
            text: str,
            speaker: dict = None,
            temperature: float = 0.1,
            repetition_penalty: float = 1.1,
            max_length = 4096,
            additional_gen_config = {},
        ) -> ModelOutput:
        input_ids = self.prepare_prompt(text, speaker)
        if self.verbose:
            logger.info(f"Input tokens: {len(input_ids)}")
            logger.info("Generating audio...")

        self.check_generation_max_length(max_length)

        output = self.model.generate(
            input_ids=input_ids,
            config=GenerationConfig(
                temperature=temperature,
                max_length=max_length,
                repetition_penalty=repetition_penalty,
                additional_gen_config=additional_gen_config,
            )
        )
        audio = self.get_audio(output)
        if self.verbose:
            logger.info("Audio generation completed")

        return ModelOutput(audio, self.audio_codec.sr)

class InterfaceEXL2(InterfaceHF):
    def __init__(
        self,
        config: EXL2ModelConfig
    ) -> None:
        self.device = torch.device(
            config.device if config.device is not None
            else "cuda" if torch.cuda.is_available()
            else "cpu"
        )
        self.config = config
        self._device = config.device
        self.languages = config.languages
        self.language = config.language
        self.verbose = config.verbose

        self.audio_codec = AudioCodec(self.device, config.wavtokenizer_model_path)
        self.prompt_processor = PromptProcessor(config.tokenizer_path, self.languages)
        self.model = EXL2Model(
            model_path=config.model_path,
            max_seq_length=config.max_seq_length,
            additional_model_config=config.additional_model_config,
        )

    def prepare_prompt(self, text: str, speaker: dict = None):
        return self.prompt_processor.get_completion_prompt(text, self.language, speaker)

    def generate(
            self,
            text: str,
            speaker: dict = None,
            temperature: float = 0.1,
            repetition_penalty: float = 1.1,
            max_length = 4096,
            additional_gen_config = {},
            additional_dynamic_generator_config = {},
        ) -> ModelOutput:
        input_ids = self.prepare_prompt(text, speaker)
        if self.verbose:
            logger.info(f"Input tokens: {len(input_ids)}")
            logger.info("Generating audio...")

        self.check_generation_max_length(max_length)

        output = self.model.generate(
            input_ids=input_ids,
            config=GenerationConfig(
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                max_length=max_length,
                additional_gen_config=additional_gen_config,
            ),
            additional_dynamic_generator_config=additional_dynamic_generator_config
        )
        audio = self.get_audio(output)
        if self.verbose:
            logger.info("Audio generation completed")

        return ModelOutput(audio, self.audio_codec.sr)



from dataclasses import dataclass, field

@dataclass
class GenerationConfig:
    text: str = None
    voice_characteristics: str = None
    speaker: dict = None

    temperature: float = 0.1
    repetition_penalty: float = 1.1
    max_length: int = 4096
    additional_gen_config: dict = field(default_factory=lambda: {})
    additional_dynamic_generator_config: dict = field(default_factory=lambda: {})

